from typing import TYPE_CHECKING, Callable, Tuple, Type, TypeVar

import jax
from jax import numpy as jnp

if TYPE_CHECKING:
    from ._base import MatrixLieGroup


def get_epsilon(dtype: jnp.dtype) -> float:
    return {
        jnp.dtype("float32"): 1e-5,
        jnp.dtype("float64"): 1e-10,
    }[dtype]


T = TypeVar("T", bound="MatrixLieGroup")


def register_lie_group(
    *,
    matrix_dim: int,
    parameters_dim: int,
    tangent_dim: int,
    space_dim: int,
) -> Callable[[Type[T]], Type[T]]:
    """Decorator for registering Lie group dataclasses.
    - Sets static dimensionality attributes
    - Makes the group hashable
    - Marks all functions for JIT compilation
    - Adds flattening/unflattening ops for use as a PyTree node

    Example:
    ```
    @register_lie_group(
        matrix_dim=2,
        parameters_dim=2,
        tangent_dim=1,
        space_dim=2,
    )
    @dataclasses.dataclass(frozen=True)
    class SO2(_base.MatrixLieGroup):
        ...
    ```
    """

    def _wrap(cls: Type[T]) -> Type[T]:
        # Register dimensions as class attributes
        cls.matrix_dim = matrix_dim
        cls.parameters_dim = parameters_dim
        cls.tangent_dim = tangent_dim
        cls.space_dim = space_dim

        # Hash based on object ID, rather than contents (arrays are not hashable)
        setattr(cls, "__hash__", object.__hash__)

        # JIT for all functions
        for f in filter(
            lambda f: not f.startswith("_") and callable(getattr(cls, f)),
            dir(cls),
        ):
            setattr(cls, f, jax.jit(getattr(cls, f)))

        # Register as a PyTree node to make JIT compilation, etc easier in Jax
        def _flatten_group(
            v: "MatrixLieGroup",
        ) -> Tuple[Tuple[jnp.ndarray, ...], Tuple]:
            """Flatten a dataclass for use as a PyTree."""
            as_dict = vars(v)
            return tuple(as_dict.values()), tuple(as_dict.keys())

        def _unflatten_group(
            treedef: Tuple, children: Tuple[jnp.ndarray, ...]
        ) -> "MatrixLieGroup":
            """Unflatten a dataclass for use as a PyTree."""
            # Treedef is names of fields
            return cls(**dict(zip(treedef, children)))  # type: ignore

        jax.tree_util.register_pytree_node(cls, _flatten_group, _unflatten_group)

        return cls

    return _wrap



import dataclasses

import numpy as onp
from jax import numpy as jnp
from overrides import overrides

from . import _base, types
from ._so3 import SO3
from ._utils import get_epsilon, register_lie_group


def _skew(omega: jnp.ndarray) -> jnp.ndarray:
    """Returns the skew-symmetric form of a length-3 vector. """

    wx, wy, wz = omega
    return jnp.array(
        [
            [0.0, -wz, wy],
            [wz, 0.0, -wx],
            [-wy, wx, 0.0],
        ]
    )


@register_lie_group(
    matrix_dim=4,
    parameters_dim=7,
    tangent_dim=6,
    space_dim=3,
)
@dataclasses.dataclass(frozen=True)
class SE3(_base.MatrixLieGroup):
    """Special Euclidean group for proper rigid transforms in 3D."""

    # SE3-specific

    xyz_wxyz: types.Vector
    """Internal parameters. Length-3 translation followed by wxyz quaternion."""

    @overrides
    def __repr__(self):
        trans = jnp.round(self.xyz_wxyz[..., :3], 5)
        quat = jnp.round(self.xyz_wxyz[..., 3:], 5)
        return f"{self.__class__.__name__}(xyz={trans}, wxyz={quat})"

    @staticmethod
    def from_rotation_and_translation(
        rotation: SO3,
        translation: types.Vector,
    ) -> "SE3":
        assert translation.shape == (3,)
        return SE3(xyz_wxyz=jnp.concatenate([translation, rotation.wxyz]))

    @property
    def rotation(self) -> SO3:
        return SO3(wxyz=self.xyz_wxyz[3:])

    @property
    def translation(self) -> types.Vector:
        return self.xyz_wxyz[:3]

    # Factory

    @staticmethod
    @overrides
    def identity() -> "SE3":
        return SE3(xyz_wxyz=onp.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]))

    @staticmethod
    @overrides
    def from_matrix(matrix: types.Matrix) -> "SE3":
        assert matrix.shape == (4, 4)
        # Currently assumes bottom row is [0, 0, 0, 1]
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            translation=matrix[:3, 3],
        )

    # Accessors

    @overrides
    def as_matrix(self) -> types.Matrix:
        return (
            jnp.eye(4)
            .at[:3, :3]
            .set(self.rotation.as_matrix())
            .at[:3, 3]
            .set(self.translation)
        )

    @property  # type: ignore
    @overrides
    def parameters(self) -> types.Vector:
        return self.xyz_wxyz

    # Operations

    @overrides
    def apply(self: "SE3", target: types.Vector) -> types.Vector:
        assert target.shape == (3,)
        return self.rotation @ target + self.translation

    @overrides
    def multiply(self: "SE3", other: "SE3") -> "SE3":
        return SE3.from_rotation_and_translation(
            rotation=self.rotation @ other.rotation,
            translation=(self.rotation @ other.translation) + self.translation,
        )

    @staticmethod
    @overrides
    def exp(tangent: types.TangentVector) -> "SE3":
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se3.hpp#L761

        # (x, y, z, omega_x, omega_y, omega_z)
        assert tangent.shape == (6,)

        rotation = SO3.exp(tangent[3:])

        theta_squared = tangent[3:] @ tangent[3:]
        theta = jnp.sqrt(theta_squared)
        skew_omega = _skew(tangent[3:])
        use_small_theta = theta < get_epsilon(theta_squared.dtype)
        V = jnp.where(
            use_small_theta,
            rotation.as_matrix(),
            (
                jnp.eye(3)
                + (1.0 - jnp.cos(theta)) / (theta_squared) * skew_omega
                + (theta - jnp.sin(theta))
                / (theta_squared * theta)
                * (skew_omega @ skew_omega)
            ),
        )

        return SE3.from_rotation_and_translation(
            rotation=rotation,
            translation=V @ tangent[:3],
        )

    @overrides
    def log(self: "SE3") -> types.TangentVector:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se3.hpp#L223
        omega = self.rotation.log()
        theta_squared = omega @ omega
        skew_omega = _skew(omega)
        theta = jnp.sqrt(theta_squared)
        half_theta = theta / 2.0
        use_small_theta = theta < get_epsilon(theta_squared.dtype)
        V_inv = jnp.where(
            use_small_theta,
            jnp.eye(3) - 0.5 * skew_omega + (skew_omega @ skew_omega) / 12.0,
            (
                jnp.eye(3)
                - 0.5 * skew_omega
                + (1.0 - theta * jnp.cos(half_theta) / (2.0 * jnp.sin(half_theta)))
                / theta_squared
                * (skew_omega @ skew_omega)
            ),
        )
        return jnp.concatenate([V_inv @ self.translation, omega])

    @overrides
    def adjoint(self: "SE3") -> types.Matrix:
        R = self.rotation.as_matrix()
        return jnp.block(
            [
                [R, _skew(self.translation) @ R],
                [jnp.zeros((3, 3)), R],
            ]
        )

    @overrides
    def inverse(self: "SE3") -> "SE3":
        R_inv = self.rotation.inverse()
        return SE3.from_rotation_and_translation(
            rotation=R_inv,
            translation=-(R_inv @ self.translation),
        )

    @overrides
    def normalize(self: "SE3") -> "SE3":
        return SE3.from_rotation_and_translation(
            rotation=self.rotation.normalize(),
            translation=self.translation,
        )



import dataclasses

import jax
from jax import numpy as jnp
from overrides import overrides

from . import _base, types
from ._so2 import SO2
from ._utils import get_epsilon, register_lie_group


@register_lie_group(
    matrix_dim=3,
    parameters_dim=4,
    tangent_dim=3,
    space_dim=2,
)
@dataclasses.dataclass(frozen=True)
class SE2(_base.MatrixLieGroup):
    """Special Euclidean group for proper rigid transforms in 2D."""

    # SE2-specific

    xy_unit_complex: types.Vector
    """Internal parameters. `(x, y, cos, sin)`."""

    @overrides
    def __repr__(self):
        xy = jnp.round(self.xy_unit_complex[..., :2], 5)
        unit_complex = jnp.round(self.xy_unit_complex[..., 2:], 5)
        return f"{self.__class__.__name__}(xy={xy}, unit_complex={unit_complex})"

    @staticmethod
    def from_xy_theta(x: float, y: float, theta: float) -> "SE2":
        cos = jnp.cos(theta)
        sin = jnp.sin(theta)
        return SE2(xy_unit_complex=jnp.array([x, y, cos, sin]))

    @staticmethod
    def from_rotation_and_translation(
        rotation: SO2,
        translation: types.Vector,
    ) -> "SE2":
        assert translation.shape == (2,)
        return SE2(
            xy_unit_complex=jnp.concatenate([translation, rotation.unit_complex])
        )

    @property
    def rotation(self) -> SO2:
        return SO2(unit_complex=self.xy_unit_complex[2:])

    @property
    def translation(self) -> types.Vector:
        return self.xy_unit_complex[:2]

    # Factory

    @staticmethod
    @overrides
    def identity() -> "SE2":
        return SE2(xy_unit_complex=jnp.array([0.0, 0.0, 1.0, 0.0]))

    @staticmethod
    @overrides
    def from_matrix(matrix: types.Matrix) -> "SE2":
        assert matrix.shape == (3, 3)
        # Currently assumes bottom row is [0, 0, 1]
        return SE2.from_rotation_and_translation(
            rotation=SO2.from_matrix(matrix[:2, :2]),
            translation=matrix[:2, 2],
        )

    # Accessors

    @property  # type: ignore
    @overrides
    def parameters(self) -> types.Vector:
        return self.xy_unit_complex

    @overrides
    def as_matrix(self) -> types.Matrix:
        x, y, cos, sin = self.xy_unit_complex
        return jnp.array(
            [
                [cos, -sin, x],
                [sin, cos, y],
                [0.0, 0.0, 1.0],
            ]
        )

    # Operations

    @overrides
    def apply(self: "SE2", target: types.Vector) -> types.Vector:
        return self.rotation @ target + self.translation

    @overrides
    def multiply(self: "SE2", other: "SE2") -> "SE2":
        # Apply rotation to both the rotation and translation terms of `other`
        xy_unit_complex = jax.vmap(self.rotation.apply)(
            other.xy_unit_complex.reshape((2, 2))
        ).flatten()

        # Apply translation
        xy_unit_complex = xy_unit_complex.at[:2].add(self.translation)

        return SE2(xy_unit_complex=xy_unit_complex)

    @staticmethod
    @overrides
    def exp(tangent: types.TangentVector) -> "SE2":
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se2.hpp#L558
        # Also see:
        # > http://ethaneade.com/lie.pdf

        assert tangent.shape == (3,)

        theta = tangent[2]
        use_taylor = jnp.abs(theta) < get_epsilon(tangent.dtype)
        theta_sq = theta ** 2
        sin_over_theta = jnp.where(
            use_taylor,
            1.0 - theta_sq / 6.0,
            jnp.sin(theta) / theta,
        )
        one_minus_cos_over_theta = jnp.where(
            use_taylor,
            0.5 * theta - theta * theta_sq / 24.0,
            (1.0 - jnp.cos(theta)) / theta,
        )

        V = jnp.array(
            [
                [sin_over_theta, -one_minus_cos_over_theta],
                [one_minus_cos_over_theta, sin_over_theta],
            ]
        )
        return SE2.from_rotation_and_translation(
            rotation=SO2.from_radians(theta),
            translation=V @ tangent[:2],
        )

    @overrides
    def log(self: "SE2") -> types.TangentVector:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/se2.hpp#L160
        # Also see:
        # > http://ethaneade.com/lie.pdf

        theta = self.rotation.log()[0]

        cos = jnp.cos(theta)
        cos_minus_one = cos - 1.0
        half_theta = theta / 2.0
        use_taylor = jnp.abs(cos_minus_one) < get_epsilon(theta.dtype)
        half_theta_over_tan_half_theta = jnp.where(
            use_taylor,
            # First-order Taylor approximation
            1.0 - (theta ** 2) / 12.0,
            # Default
            -(half_theta * jnp.sin(theta)) / cos_minus_one,
        )

        V_inv = jnp.array(
            [
                [half_theta_over_tan_half_theta, half_theta],
                [-half_theta, half_theta_over_tan_half_theta],
            ]
        )

        tangent = jnp.concatenate([V_inv @ self.translation, theta[None]])
        return tangent

    @overrides
    def adjoint(self: "SE2") -> types.Matrix:
        x, y, cos, sin = self.xy_unit_complex
        return jnp.array(
            [
                [cos, -sin, y],
                [sin, cos, -x],
                [0.0, 0.0, 1.0],
            ]
        )

    @overrides
    def inverse(self: "SE2") -> "SE2":
        R_inv = self.rotation.inverse()
        return SE2.from_rotation_and_translation(
            rotation=R_inv,
            translation=-(R_inv @ self.translation),
        )

    @overrides
    def normalize(self: "SE2") -> "SE2":
        return SE2.from_rotation_and_translation(
            rotation=self.rotation.normalize(),
            translation=self.translation,
        )



from . import manifold, types
from ._base import MatrixLieGroup
from ._se2 import SE2
from ._se3 import SE3
from ._so2 import SO2
from ._so3 import SO3
from ._utils import register_lie_group

__all__ = [
    "manifold",
    "types",
    "MatrixLieGroup",
    "SE2",
    "SO2",
    "SE3",
    "SO3",
    "types",
    "register_lie_group",
]



from typing import TypeVar, cast

import jax
from jax import numpy as jnp

from .. import types
from .._base import MatrixLieGroup
from .._se2 import SE2
from .._se3 import SE3
from .._so2 import SO2
from .._so3 import SO3

T = TypeVar("T", bound=MatrixLieGroup)


@jax.jit
def rplus(transform: T, delta: types.TangentVector) -> T:
    """Manifold right plus.

    Computes `T_wb = T_wa @ exp(delta)`.

    Args:
        transform (T): `T_wa`
        delta (types.TangentVector): `T_ab.log()`

    Returns:
        T: `T_wb`
    """
    return transform @ type(transform).exp(delta)


@jax.jit
def rplus_jacobian_parameters_wrt_delta(transform: MatrixLieGroup) -> jnp.ndarray:
    """Analytical Jacobians for `jaxlie.manifold.rplus()`, linearized around a zero
    local delta.

    Useful for on-manifold optimization.

    Equivalent to --
    ```
    def rplus_jacobian_parameters_wrt_delta(transform: MatrixLieGroup) -> jnp.ndarray:
        # Since transform objects are PyTree containers, note that `jacfwd` returns a
        # transformation object itself and that the Jacobian terms corresponding to the
        # parameters are grabbed explicitly.
        return jax.jacfwd(
            jaxlie.manifold.rplus,  # Args are (transform, delta)
            argnums=1,  # Jacobian wrt delta
        )(transform, onp.zeros(transform.tangent_dim)).parameters
    ```

    Args:
        transform (T): transform

    Returns:
        jnp.ndarray: Jacobian. Shape should be `(Group.parameters_dim, Group.tangent_dim)`.
    """
    if type(transform) is SO2:
        # Jacobian row indices: cos, sin
        # Jacobian col indices: theta

        transform_so2 = cast(SO2, transform)
        J = jnp.zeros((2, 1))

        cos, sin = transform_so2.unit_complex
        J = J.at[0].set(-sin).at[1].set(cos)

    elif type(transform) is SE2:
        # Jacobian row indices: x, y, cos, sin
        # Jacobian col indices: vx, vy, omega

        transform_se2 = cast(SE2, transform)
        J = jnp.zeros((4, 3))

        # Translation terms
        J = J.at[:2, :2].set(transform_se2.rotation.as_matrix())

        # Rotation terms
        J = J.at[2:4, 2:3].set(
            rplus_jacobian_parameters_wrt_delta(transform_se2.rotation)
        )

    elif type(transform) is SO3:
        # Jacobian row indices: qw, qx, qy, qz
        # Jacobian col indices: omega x, omega y, omega z

        transform_so3 = cast(SO3, transform)

        w, x, y, z = transform_so3.wxyz
        _unused_neg_w, neg_x, neg_y, neg_z = -transform_so3.wxyz

        J = (
            jnp.array(
                [
                    [neg_x, neg_y, neg_z],
                    [w, neg_z, y],
                    [z, w, neg_x],
                    [neg_y, x, w],
                ]
            )
            / 2.0
        )

    elif type(transform) is SE3:
        # Jacobian row indices: x, y, z, qw, qx, qy, qz
        # Jacobian col indices: vx, vy, vz, omega x, omega y, omega z

        transform_se3 = cast(SE3, transform)
        J = jnp.zeros((7, 6))

        # Translation terms
        J = J.at[:3, :3].set(transform_se3.rotation.as_matrix())

        # Rotation terms
        J = J.at[3:7, 3:6].set(
            rplus_jacobian_parameters_wrt_delta(transform_se3.rotation)
        )

    else:
        assert False, f"Unsupported type: {type(transform)}"

    assert J.shape == (transform.parameters_dim, transform.tangent_dim)
    return J


@jax.jit
def rminus(a: T, b: T) -> types.TangentVector:
    """Manifold right minus.

    Computes `delta = (T_wa.inverse() @ T_wb).log()`.

    Args:
        a (T): `T_wa`
        b (T): `T_wb`

    Returns:
        types.TangentVector: `T_ab.log()`
    """
    return (a.inverse() @ b).log()



import dataclasses

import numpy as onp
from jax import numpy as jnp
from overrides import overrides

from . import _base, types
from ._utils import get_epsilon, register_lie_group


@register_lie_group(
    matrix_dim=3,
    parameters_dim=4,
    tangent_dim=3,
    space_dim=3,
)
@dataclasses.dataclass(frozen=True)
class SO3(_base.MatrixLieGroup):
    """Special orthogonal group for 3D rotations."""

    # SO3-specific

    wxyz: types.Vector
    """Internal parameters. `(w, x, y, z)` quaternion."""

    @overrides
    def __repr__(self):
        wxyz = jnp.round(self.wxyz, 5)
        return f"{self.__class__.__name__}(wxyz={wxyz})"

    @staticmethod
    def from_x_radians(theta: float) -> "SO3":
        """Generates a x-axis rotation.

        Args:
            angle: X rotation, in radians.

        Returns:
            SO3: Output.
        """
        return SO3.exp(jnp.array([theta, 0.0, 0.0]))

    @staticmethod
    def from_y_radians(theta: float) -> "SO3":
        """Generates a y-axis rotation.

        Args:
            angle: Y rotation, in radians.

        Returns:
            SO3: Output.
        """
        return SO3.exp(jnp.array([0.0, theta, 0.0]))

    @staticmethod
    def from_z_radians(theta: float) -> "SO3":
        """Generates a z-axis rotation.

        Args:
            angle: Z rotation, in radians.

        Returns:
            SO3: Output.
        """
        return SO3.exp(jnp.array([0.0, 0.0, theta]))

    @staticmethod
    def from_rpy_radians(roll: float, pitch: float, yaw: float) -> "SO3":
        """Generates a transform from a set of Euler angles.
        Uses the ZYX mobile robot convention.

        Args:
            roll: X rotation, in radians. Applied first.
            pitch: Y rotation, in radians. Applied second.
            yaw: Z rotation, in radians. Applied last.

        Returns:
            SO3: Output.
        """
        return (
            SO3.from_z_radians(yaw)
            @ SO3.from_y_radians(pitch)
            @ SO3.from_x_radians(roll)
        )

    @staticmethod
    def from_quaternion_xyzw(xyzw: jnp.ndarray) -> "SO3":
        """Construct a rotation from an `xyzw` quaternion.

        Note that `wxyz` quaternions can be constructed using the default dataclass
        constructor.

        Args:
            xyzw (jnp.ndarray): xyzw quaternion. Shape should be (4,).

        Returns:
            SO3: Output.
        """
        assert xyzw.shape == (4,)
        return SO3(jnp.roll(xyzw, shift=1))

    def as_quaternion_xyzw(self) -> jnp.ndarray:
        """Grab parameters as xyzw quaternion."""
        return jnp.roll(self.wxyz, shift=-1)

    # Factory

    @staticmethod
    @overrides
    def identity() -> "SO3":
        return SO3(wxyz=onp.array([1.0, 0.0, 0.0, 0.0]))

    @staticmethod
    @overrides
    def from_matrix(matrix: types.Matrix, method="day") -> "SO3":
        assert matrix.shape == (3, 3)

        # Modified from:
        # > "Converting a Rotation types.Matrix to a Quaternion" from Mike Day
        # > https://d3cw3dd2w32x2b.cloudfront.net/wp-content/uploads/2015/01/matrix-to-quat.pdf

        def case0(m):
            t = 1 + m[0, 0] - m[1, 1] - m[2, 2]
            q = jnp.array([m[2, 1] - m[1, 2], t, m[1, 0] + m[0, 1], m[0, 2] + m[2, 0]])
            return t, q

        def case1(m):
            t = 1 - m[0, 0] + m[1, 1] - m[2, 2]
            q = jnp.array([m[0, 2] - m[2, 0], m[1, 0] + m[0, 1], t, m[2, 1] + m[1, 2]])
            return t, q

        def case2(m):
            t = 1 - m[0, 0] - m[1, 1] + m[2, 2]
            q = jnp.array([m[1, 0] - m[0, 1], m[0, 2] + m[2, 0], m[2, 1] + m[1, 2], t])
            return t, q

        def case3(m):
            t = 1 + m[0, 0] + m[1, 1] + m[2, 2]
            q = jnp.array([t, m[2, 1] - m[1, 2], m[0, 2] - m[2, 0], m[1, 0] - m[0, 1]])
            return t, q

        # # We can also choose to branch, but this is slower
        # t, q = jax.lax.cond(
        #     matrix[2, 2] < 0,
        #     true_fun=lambda matrix: jax.lax.cond(
        #         matrix[0, 0] > matrix[1, 1],
        #         true_fun=case0,
        #         false_fun=case1,
        #         operand=matrix,
        #     ),
        #     false_fun=lambda matrix: jax.lax.cond(
        #         matrix[0, 0] < -matrix[1, 1],
        #         true_fun=case2,
        #         false_fun=case3,
        #         operand=matrix,
        #     ),
        #     operand=matrix,
        # )

        # Compute four cases, then pick the most precise one
        # Probably worth revisiting this!
        case0_t, case0_q = case0(matrix)
        case1_t, case1_q = case1(matrix)
        case2_t, case2_q = case2(matrix)
        case3_t, case3_q = case3(matrix)

        cond0 = matrix[2, 2] < 0
        cond1 = matrix[0, 0] > matrix[1, 1]
        cond2 = matrix[0, 0] < -matrix[1, 1]

        t = jnp.where(
            cond0,
            jnp.where(cond1, case0_t, case1_t),
            jnp.where(cond2, case2_t, case3_t),
        )
        q = jnp.where(
            cond0,
            jnp.where(cond1, case0_q, case1_q),
            jnp.where(cond2, case2_q, case3_q),
        )

        return SO3(wxyz=q * 0.5 / jnp.sqrt(t))

    # Accessors

    @overrides
    def as_matrix(self) -> types.Matrix:
        norm = self.wxyz @ self.wxyz
        q = self.wxyz * jnp.sqrt(2.0 / norm)
        q = jnp.outer(q, q)
        return jnp.array(
            [
                [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0]],
                [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0]],
                [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2]],
            ]
        )

    @property  # type: ignore
    @overrides
    def parameters(self) -> types.Vector:
        return self.wxyz

    # Operations

    @overrides
    def apply(self: "SO3", target: types.Vector) -> types.Vector:
        assert target.shape == (3,)

        # Compute using quaternion multiplys
        padded_target = jnp.zeros(4).at[1:].set(target)
        return (self @ SO3(wxyz=padded_target) @ self.inverse()).wxyz[1:]

    @overrides
    def multiply(self: "SO3", other: "SO3") -> "SO3":
        w0, x0, y0, z0 = self.wxyz
        w1, x1, y1, z1 = other.wxyz
        return SO3(
            wxyz=jnp.array(
                [
                    -x0 * x1 - y0 * y1 - z0 * z1 + w0 * w1,
                    x0 * w1 + y0 * z1 - z0 * y1 + w0 * x1,
                    -x0 * z1 + y0 * w1 + z0 * x1 + w0 * y1,
                    x0 * y1 - y0 * x1 + z0 * w1 + w0 * z1,
                ]
            )
        )

    @staticmethod
    @overrides
    def exp(tangent: types.TangentVector) -> "SO3":
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L583

        assert tangent.shape == (3,)

        theta_squared = tangent @ tangent
        theta = jnp.sqrt(theta_squared)
        half_theta = 0.5 * theta
        theta_pow_4 = theta_squared * theta_squared

        use_taylor = theta < get_epsilon(tangent.dtype)
        real_factor = jnp.where(
            use_taylor,
            1.0 - theta_squared / 8.0 + theta_pow_4 / 384.0,
            jnp.cos(half_theta),
        )
        imaginary_factor = jnp.where(
            use_taylor,
            0.5 - theta_squared / 48.0 + theta_pow_4 / 3840.0,
            jnp.sin(half_theta) / theta,
        )

        return SO3(
            wxyz=jnp.concatenate(
                [
                    real_factor[None],
                    imaginary_factor * tangent,
                ]
            )
        )

    @overrides
    def log(self: "SO3") -> types.TangentVector:
        # Reference:
        # > https://github.com/strasdat/Sophus/blob/a0fe89a323e20c42d3cecb590937eb7a06b8343a/sophus/so3.hpp#L247

        w = self.wxyz[0]
        norm_sq = self.wxyz[1:] @ self.wxyz[1:]
        norm = jnp.sqrt(norm_sq)
        use_taylor = norm < get_epsilon(norm_sq.dtype)
        atan_factor = jnp.where(
            use_taylor,
            2.0 / w - 2.0 / 3.0 * norm_sq / (w ** 3),
            jnp.where(
                jnp.abs(w) < get_epsilon(w.dtype),
                jnp.where(w > 0, 1.0, -1.0) * jnp.pi / norm,
                2.0 * jnp.arctan(norm / w) / norm,
            ),
        )

        return atan_factor * self.wxyz[1:]

    @overrides
    def adjoint(self: "SO3") -> types.Matrix:
        return self.as_matrix()

    @overrides
    def inverse(self: "SO3") -> "SO3":
        # Negate complex terms
        return SO3(wxyz=self.wxyz * onp.array([1, -1, -1, -1]))

    @overrides
    def normalize(self: "SO3") -> "SO3":
        return SO3(wxyz=self.wxyz / jnp.linalg.norm(self.wxyz))



import abc
from typing import TypeVar, overload

from jax import numpy as jnp

from . import types

T = TypeVar("T", bound="MatrixLieGroup")


class MatrixLieGroup(abc.ABC):
    """Interface definition for matrix Lie groups."""

    # Class properties
    # > These will be set in `_utils.register_lie_group()`

    matrix_dim: int = 0
    """Dimension of square matrix output from `.as_matrix()`."""

    parameters_dim: int = 0
    """Dimension of underlying parameters, `.parameters`."""

    tangent_dim: int = 0
    """Dimension of tangent space."""

    space_dim: int = 0
    """Dimension of coordinates that can be transformed."""

    def __init__(self, parameters: jnp.ndarray):
        """Construct a group object from its underlying parameters."""

        # Note that this method is implicitly overriden by the dataclass decorator and
        # should _not_ be marked abstract.
        raise NotImplementedError()

    # Shared implementations

    @overload
    def __matmul__(self: T, other: T) -> T:
        ...

    @overload
    def __matmul__(self: T, other: types.Vector) -> types.Vector:
        ...

    def __matmul__(self, other):
        """Overload for the `@` operator.

        Switches between the group action (`.apply()`) and multiplication
        (`.multiply()`) based on the type of `other`.
        """
        if isinstance(other, types.Vector):
            return self.apply(target=other)
        if isinstance(other, MatrixLieGroup):
            return self.multiply(other=other)
        else:
            assert False, "Invalid argument"

    # Factory

    @staticmethod
    @abc.abstractmethod
    def identity() -> "MatrixLieGroup":
        """Returns identity element.

        Returns:
            types.Matrix: Identity.
        """

    @staticmethod
    @abc.abstractmethod
    def from_matrix(matrix: types.Matrix) -> "MatrixLieGroup":
        """Get group member from matrix representation.

        Args:
            matrix (jnp.ndarray): types.Matrix representaiton.

        Returns:
            T: Group member.
        """

    # Accessors

    @abc.abstractmethod
    def as_matrix(self) -> types.Matrix:
        """Get transformation as a matrix. Homogeneous for SE groups."""

    @property
    @abc.abstractmethod
    def parameters(self) -> types.Vector:
        """Get underlying representation."""

    # Operations

    @abc.abstractmethod
    def apply(self: T, target: types.Vector) -> types.Vector:
        """Applies the group action.

        Args:
            target (types.Vector): types.Vector to transform.

        Returns:
            types.Vector: Transformed vector.
        """

    @abc.abstractmethod
    def multiply(self: T, other: T) -> T:
        """Left-multiplies this transformations with another.

        Args:
            other (T): other

        Returns:
            T: self @ other
        """

    @staticmethod
    @abc.abstractmethod
    def exp(tangent: types.TangentVector) -> "MatrixLieGroup":
        """Computes `expm(wedge(tangent))`.

        Args:
            tangent (types.TangentVector): Input.

        Returns:
            MatrixLieGroup: Output.
        """

    @abc.abstractmethod
    def log(self: T) -> types.TangentVector:
        """Computes `vee(logm(transformation matrix))`.

        Returns:
            types.TangentVector: Output. Shape should be `(tangent_dim,)`.
        """

    @abc.abstractmethod
    def adjoint(self: T) -> types.Matrix:
        """Computes the adjoint, which transforms tangent vectors between tangent spaces.

        More precisely, for a transform `T`:
        ```
        T @ exp(omega) = exp(Adj_T @ omega) @ T
        ```

        For robotics, typically used for converting twists, wrenches, and Jacobians
        between our spatial and body representations.

        Returns:
            types.Matrix: Output. Shape should be `(tangent_dim, tangent_dim)`.
        """

    @abc.abstractmethod
    def inverse(self: T) -> T:
        """Computes the inverse of our transform.

        Returns:
            types.Matrix: Output.
        """

    @abc.abstractmethod
    def normalize(self: T) -> T:
        """Normalize/projects values and returns.

        Returns:
            T: Normalized group member.
        """



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jaxlie",
    version="0.0.2",
    description="Matrix Lie groups in Jax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/brentyi/jaxlie",
    author="brentyi",
    author_email="brentyi@berkeley.edu",
    license="MIT",
    packages=find_packages(exclude=["examples", "tests"]),
    package_data={"liejax": ["py.typed"]},
    python_requires=">=3.6",
    install_requires=[
        "jax",
        "jaxlib",
        "numpy",
        "overrides",
        "dataclasses; python_version < '3.7.0'",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import dataclasses

import numpy as onp
from jax import numpy as jnp
from overrides import overrides

from . import _base, types
from ._utils import register_lie_group


@register_lie_group(
    matrix_dim=2,
    parameters_dim=2,
    tangent_dim=1,
    space_dim=2,
)
@dataclasses.dataclass(frozen=True)
class SO2(_base.MatrixLieGroup):
    """Special orthogonal group for 2D rotations."""

    # SO2-specific

    unit_complex: types.Vector
    """Internal parameters. `(cos, sin)`."""

    @overrides
    def __repr__(self):
        unit_complex = jnp.round(self.unit_complex, 5)
        return f"{self.__class__.__name__}(unit_complex={unit_complex})"

    @staticmethod
    def from_radians(theta: float) -> "SO2":
        cos = jnp.cos(theta)
        sin = jnp.sin(theta)
        return SO2(unit_complex=jnp.array([cos, sin]))

    def to_radians(self) -> float:
        (radians,) = self.log()
        return radians

    # Factory

    @staticmethod
    @overrides
    def identity() -> "SO2":
        return SO2(unit_complex=onp.array([1.0, 0.0]))

    @staticmethod
    @overrides
    def from_matrix(matrix: types.Matrix) -> "SO2":
        assert matrix.shape == (2, 2)
        return SO2(unit_complex=matrix[:, 0])

    # Accessors

    @overrides
    def as_matrix(self) -> types.Matrix:
        cos, sin = self.unit_complex
        return jnp.array(
            [
                [cos, -sin],
                [sin, cos],
            ]
        )

    @property  # type: ignore
    @overrides
    def parameters(self) -> types.Vector:
        return self.unit_complex

    # Operations

    @overrides
    def apply(self: "SO2", target: types.Vector) -> types.Vector:
        assert target.shape == (2,)
        return self.as_matrix() @ target

    @overrides
    def multiply(self: "SO2", other: "SO2") -> "SO2":
        return SO2(unit_complex=self.as_matrix() @ other.unit_complex)

    @staticmethod
    @overrides
    def exp(tangent: types.TangentVector) -> "SO2":
        (theta,) = tangent
        cos = jnp.cos(theta)
        sin = jnp.sin(theta)
        return SO2(unit_complex=jnp.array([cos, sin]))

    @overrides
    def log(self: "SO2") -> types.TangentVector:
        return jnp.arctan2(self.unit_complex[1, None], self.unit_complex[0, None])

    @overrides
    def adjoint(self: "SO2") -> types.Matrix:
        return jnp.eye(1)

    @overrides
    def inverse(self: "SO2") -> "SO2":
        return SO2(unit_complex=self.unit_complex * onp.array([1, -1]))

    @overrides
    def normalize(self: "SO2") -> "SO2":
        return SO2(unit_complex=self.unit_complex / jnp.linalg.norm(self.unit_complex))



from jax import numpy as jnp

Matrix = jnp.ndarray
"""Type alias for `jnp.ndarray`. Should not be instantiated.

Refers to a square matrix, typically with shape `(Group.matrix_dim, Group.matrix_dim)`.
For adjoints, shape should be `(Group.tangent_dim, Group.tangent_dim)`.
"""


Vector = jnp.ndarray
"""Type alias for `jnp.ndarray`. Should not be instantiated.

Refers to a general 1D array.
"""


TangentVector = jnp.ndarray
"""Type alias for `jnp.ndarray`. Should not be instantiated.

Refers to a 1D array with shape `(Group.tangent_dim,)`.
"""


__all__ = [
    "Matrix",
    "Vector",
    "TangentVector",
]
